from marshmallow import post_load
from marshmallow_peewee import ModelSchema as MMPWModelSchema
from marshmallow_peewee import schema as marshmallow_peewee_schema

from colibris.utils import camelcase_to_underscore

DEFAULT_DATETIME_FORMAT = '%Y-%m-%dT%H:%M:%S'


class ModelSchemaOpts(marshmallow_peewee_schema.SchemaOpts):
    def __init__(self, meta, **kwargs):
        super().__init__(meta, **kwargs)

        if self.model:
            self.name = getattr(meta, 'name', camelcase_to_underscore(self.model.__name__))
            self.name_plural = getattr(meta, 'name_plural', self.name + 's')

        self.load_instance = getattr(meta, 'load_instance', False)

        self.datetimeformat = DEFAULT_DATETIME_FORMAT


class ModelSchema(MMPWModelSchema):
    OPTIONS_CLASS = ModelSchemaOpts

    @post_load
    def make_instance(self, data, **kwargs):
        if not self.opts.load_instance:
            return data

        return super().make_instance(data)



from os.path import join

from aiohttp.web_response import Response, json_response

from colibris.docs.openapi import STATIC_PATH, APISPEC_URL, STATIC_URL


async def ui_view(request):
    with open(join(STATIC_PATH, "index.html")) as f:
        content = f.read()

    content = content.replace("##API_SPEC_URL##", APISPEC_URL)
    content = content.replace("##STATIC_BASE_URL##", STATIC_URL)

    return Response(text=content, content_type="text/html")


async def apispec_view(request):
    return json_response(request.app["swagger_dict"])




import logging

from colibris.conf import settings

from .base import ABSENT
from .base import CacheBackend


DEFAULT_LIFETIME = 300  # seconds

logger = logging.getLogger(__name__)

_default_lifetime = None


def get(key, default=None):
    value = CacheBackend.get_instance().get(key)
    if value is ABSENT:
        return default

    return value


def set(key, value, lifetime=None):
    lifetime = lifetime or _default_lifetime

    CacheBackend.get_instance().set(key, value, lifetime)


def delete(key):
    CacheBackend.get_instance().delete(key)


def setup():
    global _default_lifetime

    cache_settings = dict(settings.CACHE)
    _default_lifetime = cache_settings.pop('default_lifetime', DEFAULT_LIFETIME)

    CacheBackend.configure(cache_settings)


def is_enabled():
    return CacheBackend.is_enabled()


def is_created():
    return CacheBackend.is_created()




import logging
import os

from colibris.conf import settings

from .exceptions import *
from .base import TemplateBackend


PACKAGE_TEMPLATE_PATH = 'templates'

logger = logging.getLogger(__name__)

_backend = None


def render(template_name, **context):
    return TemplateBackend.get_instance().render(template_name, context)


def render_string(template_str, **context):
    return TemplateBackend.get_instance().render_string(template_str, context)


def setup():
    template_settings = dict(settings.TEMPLATE)
    package_template_path = os.path.join(settings.PROJECT_PACKAGE_DIR, PACKAGE_TEMPLATE_PATH)

    # Prepend package template dir to path array
    template_settings['paths'] = [package_template_path] + list(template_settings.get('paths', []))

    TemplateBackend.configure(template_settings)


def is_enabled():
    return TemplateBackend.is_enabled()


def is_created():
    return TemplateBackend.is_created()




import asyncio
import logging
import redis
import rq.timeouts

from rq import Connection
from rq.worker import Worker

from . import TaskExecException
from .base import TaskQueueBackend


DEFAULT_POLL_RESULTS_INTERVAL = 1  # seconds

logger = logging.getLogger(__name__)


class RQBackend(TaskQueueBackend):
    def __init__(self, host='localhost', port=6379, db=0, password=None,
                 poll_results_interval=DEFAULT_POLL_RESULTS_INTERVAL):

        self._host = host
        self._port = port
        self._db = db
        self._password = password
        self._poll_results_interval = poll_results_interval

        self._pending_results = []
        self._poll_loop_started = False
        self._connection = None
        self._queue = None

    def _get_connection(self):
        if self._connection is None:
            logger.debug('initializing redis connection')

            self._connection = redis.Redis(host=self._host, port=self._port, db=self._db, password=self._password)

        return self._connection

    def _get_queue(self):
        if self._queue is None:
            logger.debug('initializing task queue')

            self._queue = rq.Queue(connection=self._get_connection())

        return self._queue

    async def _poll_results(self):
        self._poll_loop_started = True
        logger.debug('starting results polling loop ')

        loop = asyncio.get_event_loop()

        while loop.is_running():
            remaining_results = []
            for tup in self._pending_results:
                try:
                    result, timeout, future = tup
                    result.refresh()

                    if result.is_finished and not future.cancelled():
                        future.set_result(result.return_value)

                    elif result.is_failed:
                        exc_value = TaskExecException(result.exc_info)

                        if not future.cancelled():
                            future.set_exception(exc_value)

                    else:
                        remaining_results.append(tup)

                except Exception as e:
                    logger.error('polling results failed: %s', e, exc_info=True)

            self._pending_results = remaining_results

            await asyncio.sleep(self._poll_results_interval)

    async def execute(self, func, *args, timeout, **kwargs):
        if not self._poll_loop_started:
            _ = asyncio.ensure_future(self._poll_results())

        queue = self._get_queue()

        loop = asyncio.get_event_loop()
        future = loop.create_future()
        result = queue.enqueue(func, job_timeout=timeout, args=args, kwargs=kwargs)

        self._pending_results.append((result, timeout, future))

        return await future

    def add_worker_arguments(self, parser):
        parser.add_argument('--queue-name', help='The task queue name ("default" if unspecified)',
                            type=str, default='default')

    def run_worker(self, options):
        with Connection(self._get_connection()):
            w = Worker([options.queue_name])
            w.work()



from marshmallow import Schema, fields


def wrap_one(obj):
    return obj


def wrap_many(objs, count=None, page=0, pages=1, page_size=None):
    if page_size is None:
        page_size = len(objs)

    if count is None:
        count = len(objs)

    return {
        'results': objs,
        'count': count,
        'pages': pages,
        'page': page,
        'page_size': page_size
    }


def wrap_error(code, message, details=None):
    return {
        'code': code,
        'message': message,
        'details': details
    }


def many_envelope(schema_class):
    class SchemaWrapper(Schema):
        results = fields.Nested(schema_class, many=True)
        count = fields.Integer()
        pages = fields.Integer()
        page = fields.Integer()
        page_size = fields.Integer()

    return SchemaWrapper()




import json
import logging
import re
import traceback

from aiohttp import web

from colibris import app
from colibris import authentication
from colibris import utils
from colibris.api import BaseJSONException, envelope
from colibris.conf import settings


logger = logging.getLogger(__name__)


def _extract_request_logging_info(request):
    info = '{method} {path}'.format(method=request.method, path=request.path)
    account = authentication.get_account(request)
    if account:
        info += ' (account={})'.format(account)

    info += ': '

    return info


@web.middleware
async def handle_errors_json(request, handler):
    try:
        return await handler(request)

    except BaseJSONException as e:
        logger.warning('%s%s', _extract_request_logging_info(request), e)

        return web.json_response(envelope.wrap_error(e.code, e.message, e.details), status=e.status)

    except web.HTTPRedirection as e:
        return e

    except (web.HTTPClientError, web.HTTPServerError) as e:
        code = utils.camelcase_to_underscore(re.sub('[^a-zA-Z0-9_]', '', e.reason))
        message = e.reason
        logger.warning('%s%s', _extract_request_logging_info(request), e)

        return web.json_response(envelope.wrap_error(code, message), status=e.status)

    except json.JSONDecodeError as e:
        code = 'invalid_json'
        message = 'Invalid JSON.'
        details = str(e)
        logger.warning('%sinvalid JSON: %s', _extract_request_logging_info(request), e)

        return web.json_response(envelope.wrap_error(code, message, details), status=400)

    except app.HealthException as e:
        code = 'unhealthy'
        message = 'Service is not healthy.'
        details = str(e)
        logger.critical('%sservice is not healthy: %s', _extract_request_logging_info(request), e)

        return web.json_response(envelope.wrap_error(code, message, details), status=500)

    except Exception as e:
        code = 'server_error'
        message = 'Internal Server Error'
        details = None
        logger.error('%sinternal server error: %s', _extract_request_logging_info(request), e, exc_info=True)

        if settings.DEBUG:
            details = traceback.format_exc()

        return web.json_response(envelope.wrap_error(code, message, details), status=500)



import inspect

from aiohttp import web


class ListMixin:
    async def get(self):
        query = self.get_query()
        schema = self.get_body_schema(many=True)

        if self.pagination_class is not None:
            return await self._get_paginated_response(query, schema)

        data = schema.dump(list(query))

        return web.json_response(data)

    async def _get_paginated_response(self, query, schema):
        paginator = self.pagination_class(query, self.request)
        paginated_query = paginator.paginate_query()
        data = schema.dump(list(paginated_query))
        paginated_data = paginator.get_enveloped_data(data)

        return web.json_response(paginated_data)


class CreateMixin:
    async def post(self):
        schema = self.get_body_schema()
        data = self.get_validated_body(schema)

        result = self.perform_create(data)
        if inspect.isawaitable(result):
            instance = await result
        else:
            instance = result

        data = schema.dump(instance)

        return web.json_response(data, status=201)

    def perform_create(self, data):
        model = self.get_model()
        instance = model.create(**data)

        return instance


class RetrieveMixin:
    async def get(self):
        schema = self.get_body_schema()
        instance = await self.get_object()

        data = schema.dump(instance)

        return web.json_response(data)


class UpdateMixin:
    async def patch(self):
        return await self.update(partial=True)

    async def put(self):
        return await self.update(partial=False)

    async def update(self, partial):
        instance = await self.get_object()

        schema = self.get_body_schema(partial=partial, instance=instance)
        data = self.get_validated_body(schema)

        result = self.perform_update(data, instance)
        if inspect.isawaitable(result):
            instance = await result
        else:
            instance = result

        data = schema.dump(instance)

        return web.json_response(data)

    def perform_update(self, data, instance):
        instance.update_fields(data)
        instance.save(only=data.keys())

        return instance


class DestroyMixin:
    async def delete(self):
        instance = await self.get_object()

        result = self.perform_destroy(instance)
        if inspect.isawaitable(result):
            await result

        return web.json_response(status=204)

    def perform_destroy(self, instance):
        instance.delete_instance()

        return None




import re

from .exceptions import AuthenticationException
from .model import ModelBackend


_AUTH_HEADER = 'Authorization'
_AUTH_TOKEN_REGEX = re.compile('Bearer (.+)', re.IGNORECASE)


class ApiKeyException(AuthenticationException):
    pass


class APIKeyBackend(ModelBackend):
    def __init__(self, key_field, **kwargs):
        self.key_field = key_field

        super().__init__(**kwargs)

    def extract_auth_data(self, request):
        token = None

        auth_header = request.headers.get(_AUTH_HEADER)
        if auth_header is not None:
            m = _AUTH_TOKEN_REGEX.match(auth_header)
            if not m:
                raise ApiKeyException('invalid authorization header')

            token = m.group(1)

        if not token:
            raise ApiKeyException('missing token')

        return token

    def get_lookup_value(self, auth_data):
        return auth_data

    def get_lookup_field(self):
        return getattr(self.model, self.key_field)




import os

from colibris.conf import settings

from __packagename__ import views

# Here are some examples of routes. Just remove what you don't need.

ROUTES = [
    (r'/', views.HomeView),
    (r'/health', views.HealthView),
    (r'/users/me', views.MeView),
    (r'/users', views.UsersView),
    (r'/users/{id:\d+}', views.UserView)
]


# Add static routes, for development purposes

if settings.DEBUG:
    STATIC_ROUTES = [
        (os.path.join(settings.PROJECT_PACKAGE_DIR, 'static'), '/static'),
    ]




import os
import pathlib
import shutil


def test_start_project_default_skeleton(colibris_env):
    colibris_env.chdir_projects()
    colibris_env.run_cmd('colibris-start-project test-start-project')

    assert 'test-start-project' in os.listdir('.')
    assert 'setup.py' in os.listdir('test-start-project')
    assert 'teststartproject' in os.listdir('test-start-project')

    os.chdir('test-start-project')
    assert 'settings.py' in os.listdir('teststartproject')
    assert 'manage.py' in os.listdir('teststartproject')
    assert 'migrations' in os.listdir('teststartproject')
    assert 'tests' in os.listdir('teststartproject')


def test_start_project_custom_skeleton(colibris_env):
    default_skeleton_dir = os.path.join(colibris_env.colibris_dir, 'colibris', 'skeleton')
    custom_skeleton_dir = os.path.join(colibris_env.projects_dir, 'custom-skeleton')

    shutil.copytree(default_skeleton_dir, custom_skeleton_dir)
    pathlib.Path(custom_skeleton_dir, 'custom-file.txt').touch()

    colibris_env.chdir_projects()
    colibris_env.run_cmd('colibris-start-project test-custom-project --skeleton {skeleton}'
                         .format(skeleton=custom_skeleton_dir))

    assert 'test-custom-project' in os.listdir('.')
    assert 'testcustomproject' in os.listdir('test-custom-project')
    assert 'custom-file.txt' in os.listdir('test-custom-project')

    os.chdir('test-custom-project')



import pytest

from colibris.middleware.errors import handle_errors_json
from colibris.pagination import PageNumberPagination
from colibris.schemas import ModelSchema

from colibris import persist
from colibris.views.generic import ListCreateModelView, RetrieveUpdateDestroyModelView


class Item(persist.Model):
    name = persist.CharField()
    info = persist.CharField()


class ItemSchema(ModelSchema):
    class Meta:
        model = Item


class ItemsView(ListCreateModelView):
    body_schema_class = ItemSchema
    pagination_class = None
    model = Item


class ItemsPaginatedView(ListCreateModelView):
    body_schema_class = ItemSchema
    pagination_class = PageNumberPagination
    model = Item


class ItemView(RetrieveUpdateDestroyModelView):
    body_schema_class = ItemSchema
    model = Item


MODELS = [Item]


@pytest.fixture
def database(database_maker):
    return database_maker(models=MODELS)


@pytest.fixture
async def http_client(http_client_maker):
    return await http_client_maker(middlewares=[handle_errors_json],
                                   routes=[('/paginated-items', ItemsPaginatedView),
                                           ('/items', ItemsView),
                                           ('/items/{id}', ItemView)])


class TestList:
    async def test_get_items(self, database, http_client):
        response = await http_client.get('/items')
        assert response.status == 200

        data = await response.json()
        assert isinstance(data, list)

    async def test_get_pagination_default_page_and_size(self, database, http_client):
        item__name = 'Ligula Egestas Fermentum'
        item_info = 'Ridiculus Fermentum Quam Porta'

        Item.create(name=item__name, info=item_info)

        response = await http_client.get('/paginated-items')
        assert response.status == 200

        data = await response.json()
        assert isinstance(data, dict)
        assert 'results' in data
        assert 'count' in data
        assert 'pages' in data
        assert 'page' in data
        assert 'page_size' in data

    async def test_get_pagination_different_page(self, database, http_client):
        item_info = 'Ridiculus Fermentum Quam Porta'
        first_item_name = 'Ligula Egestas Fermentum'
        second_item_name = 'Euismod Ipsum Vulputate'
        third_item_name = 'Pellentesque Sit Venenatis'

        Item.create(name=first_item_name, info=item_info)
        Item.create(name=second_item_name, info=item_info)
        Item.create(name=third_item_name, info=item_info)

        response = await http_client.get('/paginated-items', params={'page_size': 1, 'page': 1})
        assert response.status == 200

        data = await response.json()
        assert isinstance(data, dict)
        assert data['count'] == 3
        assert data['page'] == 1
        assert data['page_size'] == 1
        assert data['results'][0]['name'] == third_item_name

        response = await http_client.get('/paginated-items', params={'page_size': 1, 'page': 2})
        assert response.status == 200
        data = await response.json()
        assert data['results'][0]['name'] == second_item_name

        response = await http_client.get('/paginated-items', params={'page_size': 1, 'page': 3})
        assert response.status == 200
        data = await response.json()
        assert data['results'][0]['name'] == first_item_name

        response = await http_client.get('/paginated-items', params={'page_size': 1, 'page': 4})
        assert response.status == 200
        data = await response.json()
        assert data['results'] == []


class TestCreate:
    async def test_create_item(self, database, http_client):
        item_name = 'Commodo Nibh'
        item_info = 'Cras Lorem Purus Etiam Venenatis'

        response = await http_client.post('/items', json={'name': item_name, 'info': item_info})

        assert response.status == 201

        item = Item.select().where(Item.name == item_name).get()

        assert item.name == item_name

    async def test_create_item_validation(self, database, http_client):
        response = await http_client.post('/items')

        assert response.status == 400
        data = await response.json()
        assert 'name' in data['details']
        assert 'info' in data['details']


class TestUpdate:
    async def test_update_item(self, database, http_client):
        item_initial_name = 'Ligula Egestas Fermentum'
        item_updated_name = 'Cursus Inceptos'
        item_info = 'Ridiculus Fermentum Quam Porta'

        item = Item.create(name=item_initial_name, info=item_info)

        response = await http_client.patch('/items/{}'.format(item.id), json={'name': item_updated_name})

        assert response.status == 200

        updated_item = Item.select().where(Item.name == item_updated_name).get()

        assert item.id == updated_item.id

    async def test_update_item_validation(self, database, http_client):
        item_initial_name = 'Nibh Lorem Amet Aenean'
        item_info = 'Ridiculus Fermentum Quam Porta'

        item = Item.create(name=item_initial_name, info=item_info)

        response = await http_client.patch('/items/{}'.format(item.id), json={'name': None})

        assert response.status == 400

        data = await response.json()
        assert 'name' in data['details']

    async def test_update_item_not_found(self, database, http_client):
        response = await http_client.patch('/items/11011', json={'name': 'Egestas Fringilla'})

        assert response.status == 404


class TestDestroy:
    async def test_delete_item(self, database, http_client):
        item_name = 'Sit Lorem'
        item_info = 'Ridiculus Fermentum Quam Porta'

        item = Item.create(name=item_name, info=item_info)

        response = await http_client.delete('/items/{}'.format(item.id))

        assert response.status == 204

        items = Item.select().where(Item.name == item_name)
        assert items.count() == 0

    async def test_delete_item_not_found(self, database, http_client):
        response = await http_client.delete('/items/111111')

        assert response.status == 404



from marshmallow import fields


class FilterMixin:
    def __init__(self, field, operation, *args, **kwargs):
        self.field = field
        self.operation = operation

        super().__init__(*args, **kwargs)


class String(FilterMixin, fields.String):
    pass


class Integer(FilterMixin, fields.Integer):
    pass


class List(FilterMixin, fields.List):
    pass


class Boolean(FilterMixin, fields.Boolean):
    pass


class DateTime(FilterMixin, fields.DateTime):
    pass


class Date(FilterMixin, fields.Date):
    pass


class Time(FilterMixin, fields.Time):
    pass


class Float(FilterMixin, fields.Float):
    pass


class Decimal(FilterMixin, fields.Decimal):
    pass



import pytest

from __packagename__ import models
from __packagename__ import constants


TEST_USER_PASSWORD = 'eyJzdWIiOiJ0ZXN0X3VzZXIifQ'


@pytest.fixture
def test_user():
    yield models.User.create(name='test_user', key=TEST_USER_PASSWORD,
                             email='testuser@example.com', role=constants.ROLE_ADMIN)




import os

from marshmallow import pre_dump, post_dump, pre_load, post_load, validates_schema
from marshmallow import fields, validate
from marshmallow import ValidationError
from marshmallow import EXCLUDE as MM_EXCLUDE
from marshmallow.schema import Schema as MMSchema, SchemaMeta as MMSchemaMeta, SchemaOpts as MMSchemaOpts

from . import ImproperlyConfigured
from . import settings
from . import logger


class ColonSeparatedStringsField(fields.Field):
    def _serialize(self, value, attr, obj, **kwargs):
        if value is None:
            return ''

        return ':'.join(value)

    def _deserialize(self, value, attr, data, **kwargs):
        if value is None:
            return None

        return value.split(':')


class SettingsSchemaOpts(MMSchemaOpts):
    def __init__(self, meta, **kwargs):
        super().__init__(meta, **kwargs)

        self.unknown = getattr(meta, 'unknown', MM_EXCLUDE)  # Ignore other variables, by default
        self.prefix = getattr(meta, 'prefix', '')
        self.sensible_fields = getattr(meta, 'sensible_fields', [])
        self.sensible_fields = set(self.prefix + f for f in self.sensible_fields)


class SettingsSchemaMeta(MMSchemaMeta):
    @classmethod
    def get_declared_fields(mcs, klass, cls_fields, inherited_fields, dict_cls):
        prefix = klass.opts.prefix
        cls_fields = [(prefix + n, v) for n, v in cls_fields]
        inherited_fields = [(prefix + n, v) for n, v in inherited_fields]

        return MMSchemaMeta.get_declared_fields(klass, cls_fields, inherited_fields, dict_cls)


class SettingsSchema(MMSchema, metaclass=SettingsSchemaMeta):
    OPTIONS_CLASS = SettingsSchemaOpts
    LOG_FORMAT = 'loading %s = "%s"'

    def _override_setting_rec(self, setting_dict, name, value):
        parts = name.split('_')
        for i in range(len(parts) - 1):
            root_name = '_'.join(parts[:i + 1])
            d = setting_dict.get(root_name)
            if not isinstance(d, dict):
                continue

            key = '_'.join(parts[i + 1:])
            self._override_setting_rec(d, key.lower(), value)

            break

        else:
            # Simply add the new setting to the setting dict
            setting_dict[name] = value

    def _override_setting(self, settings_dict, name, value):
        # Do we have the setting corresponding to the given name?
        if name in settings_dict:
            settings_dict[name] = value
            return

        # Recursively update dictionary with items
        self._override_setting_rec(settings_dict, name, value)

    def load(self, data, target_settings=None, log_format=LOG_FORMAT):
        if target_settings is None:
            target_settings = settings.__dict__

        try:
            loaded_settings = super().load(data)

        except ValidationError as e:
            # Pull the first erroneous field with its first error message to form an ImproperlyConfigured exception
            field, messages = list(e.messages.items())[0]
            message = messages[0]

            raise ImproperlyConfigured('{}: {}'.format(field, message))

        for name, value in sorted(loaded_settings.items()):
            if value is None:
                continue

            if name in self.opts.sensible_fields:
                logger.debug(log_format, name, '*****')

            else:
                logger.debug(log_format, name, value)

            self._override_setting(target_settings, name, value)

    def load_from_env(self, target_settings=None):
        self.load(os.environ, target_settings, log_format=self.LOG_FORMAT + ' from environment')



import asyncio
import inspect
import logging
import time

from aiohttp import web, hdrs
from aiohttp_apispec import setup_aiohttp_apispec

from colibris import utils
from colibris.conf import settings

logger = logging.getLogger(__name__)
middleware = []

_web_app = None
_project_app = None
_start_time = time.time()


class HealthException(Exception):
    pass


# Web App & Middleware

def _make_web_app():
    for path in settings.MIDDLEWARE:
        middleware.append(utils.import_member(path))

    return web.Application(middlewares=middleware, client_max_size=settings.MAX_REQUEST_BODY_SIZE)


async def _init_app(app):
    global _project_app

    _project_app = utils.import_module_or_none('{}.app'.format(settings.PROJECT_PACKAGE_NAME))
    if _project_app is not None:
        init = getattr(_project_app, 'init', None)
        if init:
            init(app, asyncio.get_event_loop())


# Health

async def get_health():
    if hasattr(_project_app, 'get_health'):
        h = _project_app.get_health()
        if inspect.isawaitable(h):
            h = await h

        return h

    return {
        'uptime': int(time.time() - _start_time)
    }


async def _initial_health_check(app):
    try:
        await get_health()

    except Exception as e:
        logger.critical('initial health check problem: %s', e)
        raise


# Routes

def _add_route_tuple(web_app, route):
    path, handler = route

    # aiohttp reuses the last resource if and only if two successive routes are the same,
    # which is kind of random.

    # Explicitly add resource instead of using add_route(),
    # since we want to prevent reusing the last resource.

    resource = web_app.router.add_resource(path)
    resource.add_route(hdrs.METH_ANY, handler)


def _add_static_route_tuple(web_app, route):
    fs_path, prefix = route

    web_app.router.add_static(prefix, fs_path)


def _init_routes(web_app):
    routes = utils.import_module_or_none('{}.routes'.format(settings.PROJECT_PACKAGE_NAME))
    if routes is None:
        return

    for route in getattr(routes, 'ROUTES', []):
        _add_route_tuple(web_app, route)

    for route in getattr(routes, 'STATIC_ROUTES', []):
        _add_static_route_tuple(web_app, route)


# API Docs support

def _init_docs(web_app):
    from colibris.docs.openapi.setup import setup_openapi_ui

    setup_openapi_ui(app=web_app)
    setup_aiohttp_apispec(app=web_app, title='API Documentation')


def get_web_app(force_create=False):
    global _web_app

    if _web_app is None or force_create:
        _web_app = _make_web_app()

        _init_routes(_web_app)
        _init_docs(_web_app)

        _web_app.on_startup.append(_init_app)
        _web_app.on_startup.append(_initial_health_check)

    return _web_app




import logging

from colibris.conf import settings

from .base import AuthenticationBackend
from .exceptions import *


REQUEST_ACCOUNT_KEY = 'account'


logger = logging.getLogger(__name__)


def require_authentication(required=True):
    def decorator(handler):
        handler.authentication_required = required

        return handler

    return decorator


def get_authentication_required(handler):
    return getattr(handler, 'authentication_required', None)


def authenticate(request):
    account = AuthenticationBackend.get_instance().authenticate(request)
    request[REQUEST_ACCOUNT_KEY] = account

    return account


def get_account(request):
    return request.get(REQUEST_ACCOUNT_KEY)


def login(request, account, persistent):
    logger.debug('logging in account "%s" (persistent=%s)', account, persistent)

    request[REQUEST_ACCOUNT_KEY] = account
    AuthenticationBackend.get_instance().login(request, persistent)


def logout(request):
    account = request.pop(REQUEST_ACCOUNT_KEY, None)
    if account:
        logger.debug('logging out account "%s"', account)
        AuthenticationBackend.get_instance().logout(request)


def process_request(request):
    return AuthenticationBackend.get_instance().process_request(request)


def process_response(request, response):
    return AuthenticationBackend.get_instance().process_response(request, response)


def setup():
    AuthenticationBackend.configure(settings.AUTHENTICATION)





# Patch marshmallow to provide marshmallow.compat which was removed in 3.0.0rc6.
# Current latest version of marshmallow-peewee still uses marshmallow.compat.

try:
    from marshmallow import compat

except ImportError:
    import marshmallow
    import sys
    import types

    compat = types.ModuleType('compat')
    marshmallow.compat = compat
    sys.modules['marshmallow.compat'] = compat

    def with_metaclass(meta, *bases):
        class MetaClass(meta):
            def __new__(cls, name, this_bases, d):
                return meta(name, bases, d)

        return type.__new__(MetaClass, 'temporary_class', (), {})

    compat.with_metaclass = with_metaclass
    compat.PY2 = False

# Filter apispec multiple schemas resolved to same name warning.
import warnings
warnings.filterwarnings('ignore', category=UserWarning, message=r'Multiple schemas resolved to the name.*')




import pytest

from colibris.authorization.rights import RightsBackend
from colibris import persist


USERNAME1 = 'username1'
USERNAME2 = 'username2'

RESOURCE = 'resource'
OPERATION = 'operation'
PERMISSION = '{}:{}'.format(RESOURCE, OPERATION)


class User(persist.Model):
    username = persist.CharField()


class Right(persist.Model):
    user = persist.ForeignKeyField(User)
    resource = persist.CharField()
    operation = persist.CharField()


@pytest.fixture
def database(database_maker):
    return database_maker(models=[User, Right])


@pytest.fixture
def user1(database):
    return User.create(username=USERNAME1)


@pytest.fixture
def user2(database):
    return User.create(username=USERNAME2)


@pytest.fixture
def right(user1):
    return Right.create(user=user1, resource=RESOURCE, operation=OPERATION)


@pytest.fixture
def backend():
    return RightsBackend(model=Right, account_field='user',
                         resource_field='resource', operation_field='operation')


def test_allowed(backend, user1, right):
    assert backend.get_actual_permissions(account=user1, method='GET', path='/') == {PERMISSION}


def test_forbidden(backend, user2, right):
    assert backend.get_actual_permissions(account=user2, method='GET', path='/') == set()




import sys

from colibris import conf
from colibris import utils
from colibris.conf import settings

from .base import BaseCommand

from . import db
from . import runserver
from . import runworker
from . import shell
from . import test


_command = None  # Currently running command


def gather_all_commands():
    global ALL_COMMANDS

    # Import any project-specific commands
    utils.import_module_or_none('{}.commands'.format(settings.PROJECT_PACKAGE_NAME))

    ALL_COMMANDS = {c.get_name(): c for c in BaseCommand.__subclasses__()}


def show_commands_usage():
    cmds = '\n  '.join([c.get_name() for c in ALL_COMMANDS.values()])
    usage = '{prog} <command> [args]\n\n' \
            'available commands:\n  {cmds}\n\n' \
            'use {prog} <command> -h for help'

    usage = usage.format(cmds=cmds, prog='manage.py')

    print(usage)


def get_command():
    return _command


def main():
    global _command

    conf.setup()

    gather_all_commands()

    args = sys.argv
    if len(args) < 2 or args[1] not in ALL_COMMANDS:
        show_commands_usage()
        sys.exit(1)

    command_class = ALL_COMMANDS[sys.argv[1]]
    _command = command_class(args[2:])
    exit_code = _command.run() or 0

    sys.exit(exit_code)




from colibris.conf.backends import BackendMixin


class TaskQueueBackend(BackendMixin):
    async def execute(self, func, *args, timeout, **kwargs):
        raise NotImplementedError

    def add_worker_arguments(self, parser):
        pass

    def run_worker(self, options):
        raise NotImplementedError




import jwt
import re
import time

from .cookie import CookieBackendMixin
from .exceptions import AuthenticationException
from .model import ModelBackend


_AUTH_HEADER = 'Authorization'
_AUTH_TOKEN_REGEX = re.compile('Bearer (.+)', re.IGNORECASE)

DEFAULT_VALIDITY_SECONDS = 3600 * 24 * 30
JWT_ALG = 'HS256'


class JWTException(AuthenticationException):
    pass


class JWTBackend(ModelBackend, CookieBackendMixin):
    def __init__(self, identity_field, secret_field, identity_claim='sub', **kwargs):
        self.identity_claim = identity_claim
        self.identity_field = identity_field
        self.secret_field = secret_field

        super().__init__(**kwargs)

    def extract_auth_data(self, request):
        token = None

        # First look for auth header
        auth_header = request.headers.get(_AUTH_HEADER)
        if auth_header is not None:
            m = _AUTH_TOKEN_REGEX.match(auth_header)
            if not m:
                raise JWTException('invalid authorization header')

            token = m.group(1)

        # Then look in cookies
        elif self.cookie_name:
            token = request.cookies.get(self.cookie_name)

        if not token:
            raise JWTException('missing token')

        try:
            jwt_decoded = jwt.decode(token, verify=False)

        except jwt.DecodeError:
            raise JWTException('invalid token')

        identity = jwt_decoded.get(self.identity_claim)
        if identity is None:
            raise JWTException('missing identity claim')

        return {
            'identity': identity,
            'token': token
        }

    def get_lookup_value(self, auth_data):
        return auth_data['identity']

    def get_lookup_field(self):
        return getattr(self.model, self.identity_field)

    def verify_identity(self, request, account, auth_data):
        secret = self.get_secret(account)

        try:
            jwt.decode(auth_data['token'], key=secret, verify=True, algorithms=[JWT_ALG])

        except jwt.InvalidSignatureError:
            raise JWTException('invalid signature')

        except jwt.InvalidAlgorithmError:
            raise JWTException("invalid algorithm")

        if self.is_csrf_enabled():
            self.verify_csrf(request, account)

    def prepare_login_response(self, response, account, persistent):
        self.cookie_login(response, persistent, self.build_jwt(account))

        return response

    def prepare_logout_response(self, response):
        self.cookie_logout(response)

        return response

    def build_jwt(self, account):
        now = int(time.time())
        token_claims = {
            'sub': getattr(account, self.identity_field),
            'iat': now,
            'exp': now + self.validity_seconds
        }

        return jwt.encode(algorithm=JWT_ALG, payload=token_claims, key=self.get_secret(account)).decode()

    def process_response(self, request, response):
        response = super().process_response(request, response)

        if self.is_csrf_enabled():
            self.set_csrf(request, response)

        return response

    def get_secret(self, account):
        return getattr(account, self.secret_field)




import json
import os
import requests
import time


def test_manage_make_migrations(test_project):
    test_project.run_cmd('testproject/manage.py db create .models')

    assert '0001_create_table_user.py' in os.listdir('testproject/migrations')


def test_manage_migrate(test_project):
    test_project.run_cmd('testproject/manage.py db create .models')
    test_project.run_cmd('testproject/manage.py db upgrade')
    test_project.run_cmd('sqlite3 testproject.db .tables | grep user')


def test_manage_runserver(test_project):
    test_project.run_cmd('testproject/manage.py db create .models')
    test_project.run_cmd('testproject/manage.py db upgrade')

    process = test_project.run_popen('testproject/manage.py runserver')
    time.sleep(2)  # Wait for the server to start

    response = requests.get('http://localhost:8888/health')

    try:
        assert response.status_code == 200
        assert json.loads(response.text) == 'healthy'

    finally:
        process.kill()
        time.sleep(2)  # Wait for the server to stop




import os
import pytest
import subprocess
import sys
import tempfile


_colibris_env = None


class ColibrisEnv:
    def __init__(self):
        self.venvs_tmp_dir = tempfile.TemporaryDirectory()
        self.venvs_dir = self.venvs_tmp_dir.name
        self.venv_dir = os.path.join(self.venvs_dir, 'test_venv')

        self.projects_tmp_dir = tempfile.TemporaryDirectory()
        self.projects_dir = self.projects_tmp_dir.name

        self.colibris_dir = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))

        # Prepare virtual env & install colibris
        self.make_virtual_env()
        dist = self.make_dist()
        self.run_cmd('pip install {}'.format(dist))

    def __del__(self):
        self.venvs_tmp_dir.cleanup()
        self.projects_tmp_dir.cleanup()

    def make_virtual_env(self):
        subprocess.check_call('pip install virtualenv', shell=True)
        subprocess.check_call('python -m virtualenv {}'.format(self.venv_dir), shell=True)

    def make_dist(self):
        prev_dir = os.getcwd()
        os.chdir(self.colibris_dir)
        subprocess.check_call('rm -rf dist build *.egg', shell=True)
        subprocess.check_call('{} setup.py sdist'.format(sys.executable), shell=True)
        os.chdir(prev_dir)

        return os.path.join(self.colibris_dir, 'dist', os.listdir(self.colibris_dir + '/dist')[0])

    def chdir_projects(self):
        os.chdir(self.projects_dir)

    def ensure_req(self, req):
        self.run_cmd('pip install {}'.format(req))

    def run_cmd(self, cmd):
        subprocess.check_call('. {}/bin/activate && {}'.format(self.venv_dir, cmd), shell=True)

    def run_popen(self, cmd, **kwargs):
        return subprocess.Popen('. {}/bin/activate && {}'.format(self.venv_dir, cmd), shell=True, **kwargs)


@pytest.fixture
def colibris_env():
    global _colibris_env

    # Reuse the same env for all tests to reduce testing time
    if _colibris_env is None:
        _colibris_env = ColibrisEnv()

    return _colibris_env


@pytest.fixture
def test_project(colibris_env):
    initial_dir = os.getcwd()
    colibris_env.chdir_projects()

    colibris_env.ensure_req('pyjwt')
    colibris_env.ensure_req('pytest')

    colibris_env.run_cmd('colibris-start-project test-project'.format(colibris_env.colibris_dir))
    os.chdir('test-project')

    yield colibris_env

    colibris_env.chdir_projects()
    colibris_env.run_cmd('rm -rf test-project')
    os.chdir(initial_dir)



from . import fixtures


async def test_unauthenticated(web_app_client):
    resp = await web_app_client.get('/users')
    assert resp.status == 401

    j = await resp.json()
    assert j['code'] == 'unauthenticated'


async def test_authenticated(web_app_client, test_user):
    token = fixtures.TEST_USER_PASSWORD
    resp = await web_app_client.get('/users/me', headers={'Authorization': 'Bearer {}'.format(token)})
    assert resp.status == 200




import sys

from peewee_moves import cli_command

from colibris import persist

from .base import BaseCommand


class DBCommand(BaseCommand):

    def parse_arguments(self, parser, args):
        pass

    def execute(self, options):
        # proxy to peewee moves cli
        cli_command(
            default_map={
                "directory": persist.get_migrations_dir(),
                "database": persist.get_database,
            },
            args=sys.argv[2:],
        )




import async_timeout
import asyncio
import logging

from colibris.conf import settings

from .exceptions import *
from .exceptions import TimeoutException
from .base import TaskQueueBackend


DEFAULT_TIMEOUT = 300  # seconds

logger = logging.getLogger(__name__)


async def execute(func, *args, timeout=DEFAULT_TIMEOUT, **kwargs):
    try:
        with async_timeout.timeout(timeout):
            return await TaskQueueBackend.get_instance().execute(func, *args, timeout=timeout, **kwargs)

    except asyncio.TimeoutError:
        raise TimeoutException(timeout)


def add_worker_arguments(parser):
    TaskQueueBackend.get_instance().add_worker_arguments(parser)


def run_worker(options):
    TaskQueueBackend.get_instance().run_worker(options)


def setup():
    TaskQueueBackend.configure(settings.TASK_QUEUE)


def is_enabled():
    return TaskQueueBackend.is_enabled()


def is_created():
    return TaskQueueBackend.is_created()




import logging
import os

from peewee import *
try:
    from playhouse.postgres_ext import *

except ImportError:
    pass

from colibris.conf import settings

from .backends import DatabaseBackend, PostgreSQLBackend, MySQLBackend, SQLiteBackend
from .models import Model, set_database


_PEEWEE_DB_PARAMS_MAPPING = {
    'name': 'database',
    'username': 'user'
}

logger = logging.getLogger(__name__)


def get_migrations_dir():
    # Migrations live in the project's root package
    return os.path.join(settings.PROJECT_PACKAGE_DIR, 'migrations')


def get_database():
    return DatabaseBackend.get_instance()


def connectivity_check():
    # run a dummy SQL statement to check connectivity with DB server

    try:
        list(get_database().execute_sql("select 'dummy'"))
        return True

    except Exception:
        return False


def setup():
    db_settings = dict(settings.DATABASE)
    db_settings.setdefault('autorollback', True)

    create = db_settings.pop('create', False)

    # Translate settings into whatever peewee prefers
    for param, pparam in _PEEWEE_DB_PARAMS_MAPPING.items():
        if param in db_settings:
            db_settings[pparam] = db_settings.pop(param)

    DatabaseBackend.configure(db_settings)

    if DatabaseBackend.is_enabled():
        database = get_database()
        if create:
            logger.debug('creating db')
            try:
                database.create()

            except Exception as e:
                logger.error('db creation failed: %s', e, exc_info=True)

        database.connect()
        logger.debug('db connection initialized')

        set_database(get_database())


def is_enabled():
    return DatabaseBackend.is_enabled()


def is_created():
    return DatabaseBackend.is_created()



from colibris.schemas import ModelSchema
from colibris.schemas import fields, validate, pre_load

from __packagename__ import models


# Here is a schema example. Just remove it if you don't need it.

class UserSchema(ModelSchema):
    email = fields.String(validate=[validate.Email(error='Invalid email address.'),
                                    validate.Length(max=128)])

    @pre_load
    def process_input(self, data, **kwargs):
        if 'email' in data:
            data['email'] = data['email'].lower().strip()

        return data

    class Meta:
        model = models.User
        name = 'user'
        name_plural = 'users'




from .base import AuthorizationBackend


# A permission is defined as the role of an account.
#
# The optional ordering defines a relation between roles; this allows specifying e.g. "admin" as required permission
# and assuming that it also has "regular" and "viewonly" permissions, without explicitly requiring them.

class RoleBackend(AuthorizationBackend):
    def __init__(self, role_field, order=None, **kwargs):
        self.role_field = role_field
        self.order = order or []  # lower first, e.g. ["readonly", "regular", "admin"]

        super().__init__(**kwargs)

    def get_role(self, account):
        return getattr(account, self.role_field)

    def get_actual_permissions(self, account, method, path):
        role = self.get_role(account)
        actual_permissions = {role}

        try:
            index = self.order.index(role)

        except ValueError:
            index = 0

        actual_permissions.update(self.order[:index])

        return actual_permissions




from jinja2 import Environment
from jinja2 import FileSystemLoader
from jinja2 import TemplateNotFound, TemplateSyntaxError, TemplateError
from jinja2 import select_autoescape

from colibris import utils

from . import base
from . import exceptions


class Jinja2Backend(base.TemplateBackend):
    def __init__(self, paths, extensions=None, translations=None):
        super().__init__(paths)

        extensions = extensions or []
        if translations:
            extensions.append('jinja2.ext.i18n')

        self.loader = FileSystemLoader(paths)
        self.env = Environment(loader=self.loader, autoescape=select_autoescape(),
                               extensions=extensions)

        if translations:
            translations = utils.import_member(translations)
            self.env.install_gettext_translations(translations, newstyle=True)

    def render(self, template_name, context):
        try:
            template = self.env.get_template(template_name)
            return template.render(context)

        except TemplateNotFound:
            raise exceptions.TemplateNotFound(template_name)

        except TemplateSyntaxError as e:
            raise exceptions.TemplateSyntaxError(str(e))

        except TemplateError as e:
            raise exceptions.TemplateException(str(e))

    def render_string(self, template_str, context):
        try:
            template = self.env.from_string(template_str)
            return template.render(context)

        except TemplateSyntaxError as e:
            raise exceptions.TemplateSyntaxError(str(e))

        except TemplateError as e:
            raise exceptions.TemplateException(str(e))
